import json
from os import environ
import tempfile

import sst


with tempfile.NamedTemporaryFile() as temp_fp:

    sst.setProgramOption("stopAtCycle", "25us")

    main_comp = sst.Component("unit", "verilogsock.unit")
    main_comp.addParams(
        {
            "clock": "1MHz",
        }
    )

    ram_comp = sst.Component("ram", "verilogsock.ram")

    proc = ""
    with open("config-cmd.json") as cmd_file:
        data = json.load(cmd_file)
    for k, v in data.items():
        if k != "FILE":
            environ[k] = v
        else:
            proc = v

    ram_comp.addParams(
        {
            "proc": proc,
            "ipc_port": temp_fp.name,
            "clock": "1MHz",
        }
    )

    sst.Link("ram_din").connect(
        (ram_comp, "ram_din", "1ps"),
        (main_comp, "ram_din", "1ps"),
    )
    sst.Link("ram_dout").connect(
        (ram_comp, "ram_dout", "1ps"),
        (main_comp, "ram_dout", "1ps"),
    )



from typing import Literal

from .. import HardwareDescriptionLanguage

class Verilog(HardwareDescriptionLanguage):
    def __init__(
        self,
        module_name: str,
        lib: str,
        ipc: Literal["sock", "zmq"] = ...,
        module_dir: str = ...,
        lib_dir: str = ...,
        desc: str = ...,
    ) -> None: ...
    def _compute_signal_buffer_len(
        self, signal_type: str, signal_len: int
    ) -> int: ...
    def _get_driver_inputs(self) -> str: ...
    def _get_driver_outputs(self) -> str: ...
    def _get_driver_defs(self) -> dict[str, str]: ...
    def _generate_extra_files(self) -> None: ...



import pathlib
import subprocess
import sysconfig


class _Commands:
    @staticmethod
    def capture_output(cmd):

        return cmd.stdout.decode("utf-8")[:-1]

    def run_cmd(self, cmd, stdout=subprocess.PIPE):

        return subprocess.run(cmd, stdout=stdout)

    def __get_sst_config(self):

        sst_config = self.run_cmd(["which", "sst-config"])
        return self.capture_output(sst_config)

    def __get_prefix(self, sst_config):

        prefix = self.run_cmd([sst_config, "--prefix"])
        return pathlib.Path(self.capture_output(prefix))

    def __get_include_dir(self, sst_config):

        include_dir = self.run_cmd([sst_config, "--includedir"])
        return pathlib.Path(self.capture_output(include_dir))

    def get_dest_dir(self):

        sst_config = self.__get_sst_config()
        prefix = self.__get_prefix(sst_config)
        include_dir = self.__get_include_dir(sst_config)

        return prefix / include_dir.name / "sst" / "sit"


class LibraryManager:
    def __init__(self):

        self.__cmd = _Commands()
        self.__dest_dir = self.__cmd.get_dest_dir()

    def is_installed(self):

        return self.__dest_dir.exists()

    def get_dest_dir(self):

        return self.__dest_dir

    def install(self):

        lib_dir = pathlib.Path(sysconfig.get_paths()["purelib"]) / "sit/cpp"
        try:
            self.__cmd.run_cmd(["cmake", lib_dir], stdout=None)

        except FileNotFoundError:
            print("CMake not found")
            exit(-1)

        self.__cmd.run_cmd(["sudo", "make", "install"], stdout=None)



import pyrtl

# Define RAM parameters
depth = 256  # Number of memory locations (2^8 for 8-bit address)
width = 8  # Data width (8 bits)

# Create PyRTL signals for inputs and outputs
address = pyrtl.Input(bitwidth=8, name="address")  # 8-bit address
data_in = pyrtl.Input(bitwidth=8, name="data_in")  # 8-bit data input
cs = pyrtl.Input(bitwidth=1, name="cs")  # Chip select (active low)
oe = pyrtl.Input(bitwidth=1, name="oe")  # Output enable (active low)
we = pyrtl.Input(bitwidth=1, name="we")  # Write enable (active low)
data_out = pyrtl.Output(bitwidth=8, name="data_out")  # 8-bit data output

# Create a synchronous RAM block
ram = pyrtl.MemBlock(bitwidth=8, addrwidth=8, name="ram", asynchronous=False)

# Define the RAM read and write logic
with pyrtl.conditional_assignment:
    with cs:
        with oe:
            data_out <<= ram[address]
        with we:
            ram[address] <<= data_in

# Simulate the PyRTL design
sim_trace = pyrtl.SimulationTrace()
sim = pyrtl.Simulation(tracer=sim_trace)



from .render import TemplateRenderer

LINK_WRAPPER_STR = """{{module_name}}_link = new LinkWrapper(&m_keep_send, &m_keep_recv);
{{module_name}}_link->set_din_link(configureLink("{{module_name}}_din"));
{{module_name}}_link->set_dout_link(configureLink(
    "{{module_name}}_dout",
    new SST::Event::Handler<{{lib}}>(this, &{{lib}}::{{module_name}})
));
{{module_name}}_link->set_buffer_lengths({{buf_lens}});
"""

CONNECT_COMPS_FUNC_STR = """
def connect_comps(comp, comp_name, main_comp_name):
    sst.Link(main_comp_name + "_din").connect(
        (comp, comp_name + "_din", LINK_DELAY),
        ({{module_name}}_main, main_comp_name + "_din", LINK_DELAY),
    )
    sst.Link(main_comp_name + "_dout").connect(
        (comp, comp_name + "_dout", LINK_DELAY),
        ({{module_name}}_main, main_comp_name + "_dout", LINK_DELAY),
    )
"""


class ConfigFile:
    def generate_linkwrapper_binding(self, module_name, lib, input_ports):

        return TemplateRenderer().render(
            LINK_WRAPPER_STR,
            {
                "module_name": module_name,
                "lib": lib,
                "buf_lens": ", ".join(str(i["len"]) for i in input_ports),
            },
        )

    def generate_comps_func(self, module_name):

        return TemplateRenderer().render(
            CONNECT_COMPS_FUNC_STR, {"module_name": module_name}
        )



from sit import SIT

if __name__ == "__main__":

    config = {
        "hdl": "systemc",
        "config": {
            "module_name": "ram",
            "lib": "systemcsock",
            "desc": "Demonstration of a SystemC hardware simulation in SST",
            "module_dir": "../tests/systemc/",
        },
        "extra": {
            "width_macros": {
                "ADDR_WIDTH": 8,
                "DATA_WIDTH": 8,
            }
        },
        "ports": {
            "input": [
                {"name": "address", "type": "sc_uint<ADDR_WIDTH>", "len": 9090},
                {"name": "cs", "type": "bool", "len": -1},
                {"name": "we", "type": "bool", "len": -1},
                {"name": "oe", "type": "bool", "len": -1},
                {"name": "data_in", "type": "sc_uint<DATA_WIDTH>", "len": -1},
            ],
            "output": [
                {"name": "data_out", "type": "sc_bv<DATA_WIDTH>", "len": -1}
            ],
        },
    }

    sit_obj = SIT(config)
    sit_obj.generate_boilerplate()



import re


class TemplateRenderer:
    def __init__(self):

        self.variable_re = re.compile(
            r"\{\{(.*?)\}\}", re.IGNORECASE | re.MULTILINE
        )

    def render(self, template, mapping):

        return self.variable_re.sub(
            lambda m: str(mapping.get(m.group(1), m.group(0))), template
        )



import argparse
import pkg_resources
from pprint import pprint

from .configfile import ConfigFile
from .exceptions import ConfigException
from .hdl import pyrtl, systemc, verilog
from .libmgmt import LibraryManager


class SIT:
    def __init__(self, config_data):

        self.__config_data = config_data
        self.__instance: pyrtl.PyRTL | systemc.SystemC | verilog.Verilog
        self.__config_file: ConfigFile

    def generate_boilerplate(self):

        self.__set_hdl_str(self.__config_data["hdl"])

        if extra_configs := self.__config_data.get("extra", None):

            if fixed_width_float_output := extra_configs.get(
                "fixed_width_float_output", None
            ):
                self.__instance.fixed_width_float_output(
                    fixed_width_float_output
                )

            if disable_runtime_warnings := extra_configs.get(
                "disable_runtime_warnings", None
            ):
                self.__instance.disable_runtime_warnings(
                    disable_runtime_warnings
                )

            if width_macros := extra_configs.get("width_macros", None):
                self.__instance.set_width_macros(width_macros)

        self.__instance.set_ports(self.__config_data["ports"])
        self.__instance.generate_boilerplate()
        self.dump_summary()

    def dump_summary(self):

        summary = {
            "module name": self.__config_data["config"]["module_name"],
            "hdl": self.__config_data["hdl"],
            "generated filepaths": {
                "driver": self.__instance.paths.get_gen("driver"),
                "component": self.__instance.paths.get_gen("comp"),
            },
            "ports": {
                "input": [],
                "output": [],
                "inout": [],
            },
            "buffer sizes": {
                "driver": self.__instance.driver_buf_size,
                "component": self.__instance.comp_buf_size,
            },
        }

        for port_type in self.__config_data["ports"]:
            if self.__config_data["ports"][port_type]:
                for port in self.__config_data["ports"][port_type]:
                    summary["ports"][port_type].append(
                        {
                            "name": port["name"],
                            "data type": port["type"],
                            "buffer length": port["len"],
                        }
                    )

        pprint(summary, sort_dicts=False)

    def generate_config_file(self):

        self.__config_file = ConfigFile()
        print(
            self.__config_file.generate_linkwrapper_binding(
                self.__config_data["config"]["module_name"],
                self.__config_data["config"]["lib"],
                self.__instance._get_input_ports(),
            )
        )

    def set_config_data(self, config_data):

        self.__config_data = config_data

    def __set_hdl_str(self, hdl_str):

        match hdl_str.lower():

            case "pyrtl":
                self.__instance = pyrtl.PyRTL(**self.__config_data["config"])

            case "systemc":
                self.__instance = systemc.SystemC(
                    **self.__config_data["config"]
                )

            case "verilog":
                self.__instance = verilog.Verilog(
                    **self.__config_data["config"]
                )

            case _:
                raise ConfigException(
                    f'"{hdl_str}" is not supported as a hardware description language'
                )


def run():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=pkg_resources.get_distribution("sst_it").version,
    )
    parser.add_argument(
        "-i",
        "--install",
        default=False,
        action="store_true",
        help="install the SIT library",
    )

    args = vars(parser.parse_args())

    if args["install"]:

        lib = LibraryManager()
        if lib.is_installed():
            print(f"SIT is already installed at '{lib.get_dest_dir()}'")
            if input("Reinstall library? [y/N] ").lower() == "y":
                lib.install()

        else:
            lib.install()



"""Implementation of the PyRTL class

This class inherits from the HardwareDescriptionLanguage base class and implements its own methods of
parsing, modifying and generating boilerplate code for its specific paradigms.
"""
from .. import HardwareDescriptionLanguage


class PyRTL(HardwareDescriptionLanguage):
    def __init__(
        self,
        module_name,
        lib,
        ipc="sock",
        module_dir="",
        lib_dir="",
        desc="",
    ):
        """Constructor for PyRTL HardwareDescriptionLanguage.

        Parameters:
        -----------
        ipc : str (options: "sock", "zmq")
            method of IPC
        module : str
            SST element component and HDL module name
        lib : str
            SST element library name
        module_dir : str (default: "")
            directory of HDL module
        lib_dir : str (default: "")
            directory of HardwareDescriptionLanguage library
        desc : str (default: "")
            description of the SST model
        driver_template_path : str (default: "")
            path to the black box-driver boilerplate
        component_template_path : str (default: "")
            path to the black box-model boilerplate
        """
        super().__init__(
            ipc=ipc,
            module_name=module_name,
            lib=lib,
            module_dir=module_dir,
            lib_dir=lib_dir,
            desc=desc,
        )

        if self.ipc == "sock":

            # driver attributes
            self.driver_ipc = "socket"
            self.driver_bind = (
                "_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)"
            )
            self.send = "sendall"
            self.connect = "connect"

        elif self.ipc == "zmq":

            # driver attributes
            self.driver_ipc = "zmq"
            self.driver_bind = """context = zmq.Context()
_sock = context.socket(zmq.REQ)"""
            self.send = "send"
            self.connect = "bind"

        self.exec_cmd = """m_output.verbose(
                CALL_INFO, 1, 0, "\\033[35m[FORK] %s\\033[0m\\n\", cmd.c_str()
            );
            char* args[8];
            int i = 0;
            args[i] = std::strtok(&cmd[0u], " ");

            while (args[i] != nullptr) {
                args[++i] = strtok(nullptr, " ");
            }
            args[i] = nullptr;"""

        self.paths.set_driver_path(f"{self.module_name}_driver.py")
        self.paths.set_comp_path(f"{self.module_name}_comp.cpp")

    def _compute_signal_buffer_len(self, signal_type, signal_len):
        """Parse the type and computes its width from the signal

        Parameters:
        -----------
        signal : str
            signal definition

        Returns:
        --------
        int
            signal width
        """
        return self._get_num_digits(signal_len)

    def _get_driver_outputs(self):
        """Generate output bindings for both the components in the black box

        Returns:
        --------
        str
            snippet of code representing output bindings
        """
        return self._sig_fmt(
            fmt="str({module_name}.sim.inspect({module_name}.{sig})).encode()",
            split_func=lambda x: {
                "module_name": self.module_name,
                "sig": x["name"],
            },
            array=self._get_output_ports(),
            delim=" +\n" + " " * 8,
        )

    def _get_driver_inputs(self):
        """Generate input bindings for the driver.

        Returns:
        --------
        str
            snippet of code representing input bindings
        """
        fmt = '"{sig}": int(signal[{sp}:{sl}]),'
        start_pos = 0
        clock_fmt = '"{sig}": int(signal[{sp}:{sl}]) % 2,'
        driver_inputs = []

        for input_port in self._get_input_ports():
            driver_inputs.append(
                fmt.format(
                    sp=start_pos,
                    sl=str(input_port["len"] + start_pos),
                    sig=input_port["name"],
                )
            )
            start_pos += input_port["len"]

        if self.ports["clock"]:
            for clock_port in self.ports["clock"]:
                driver_inputs.append(
                    clock_fmt.format(
                        sp=start_pos,
                        sl=str(input_port["len"] + start_pos),
                        sig=clock_port["name"],
                    )
                )
                start_pos += int(clock_port["len"])

        self.driver_buf_size = start_pos + 1
        return ("\n" + " " * 8).join(driver_inputs)

    def _get_driver_defs(self):
        """Map definitions for the PyRTL driver format string

        Returns:
        --------
        dict(str:str)
            format mapping of template PyRTL driver string
        """
        return {
            "ipc": self.driver_ipc,
            "driver_bind": self.driver_bind,
            "connect": self.connect,
            "send": self.send,
            "module_dir": self.paths.get_module_dir().resolve(),
            "module_name": self.module_name,
            "buf_size": self.driver_buf_size,
        }



from pathlib import Path
import tempfile

import sst

BASE_PATH = Path.cwd()

with tempfile.NamedTemporaryFile() as tfp:

    sst.setProgramOption("stopAtCycle", "25us")

    main_comp = sst.Component("unit", "systemcsock.unit")
    main_comp.addParams(
        {
            "clock": "1MHz",
        }
    )

    ram_comp = sst.Component("ram", "systemcsock.ram")

    ram_comp.addParams(
        {
            "proc": BASE_PATH / "ram.o",
            "ipc_port": tfp.name,
            "clock": "1MHz",
        }
    )

    sst.Link("ram_din").connect(
        (ram_comp, "ram_din", "1ps"),
        (main_comp, "ram_din", "1ps"),
    )
    sst.Link("ram_dout").connect(
        (ram_comp, "ram_dout", "1ps"),
        (main_comp, "ram_dout", "1ps"),
    )



from sit import SIT

if __name__ == "__main__":

    config = {
        "hdl": "pyrtl",
        "config": {
            "module_name": "ram",
            "lib": "pyrtlsock",
            "desc": "Demonstration of a PyRTL hardware simulation in SST",
            "module_dir": "../tests/pyrtl/",
        },
        "ports": {
            "input": [
                {"name": "address", "type": "int", "len": 8},
                {"name": "cs", "type": "int", "len": 1},
                {"name": "we", "type": "int", "len": 1},
                {"name": "oe", "type": "int", "len": 1},
                {"name": "data_in", "type": "int", "len": 8},
            ],
            "output": [{"name": "data_out", "type": "bit", "len": 8}],
        },
    }

    sit_obj = SIT(config)
    sit_obj.generate_boilerplate()



import os

class Paths:
    def __init__(self, hdl_str: str, module_dir_path: str) -> None: ...
    def get_meta(self, key: str) -> os.PathLike: ...
    def get_template(self, key: str) -> os.PathLike: ...
    def get_gen(self, key: str) -> os.PathLike: ...
    def get_paths(self) -> dict[str, os.PathLike]: ...
    def set_template_paths(
        self,
        dir: os.PathLike = ...,
        driver: os.PathLike = ...,
        comp: os.PathLike = ...,
    ) -> None: ...
    def set_gen_paths(
        self,
        dir: os.PathLike = ...,
        driver: os.PathLike = ...,
        comp: os.PathLike = ...,
    ) -> None: ...
    def set_driver_path(self, driver_file: str) -> None: ...
    def set_comp_path(self, comp_file: str) -> None: ...
    def set_extra_file_paths(
        self, templates: dict[str, str], gens: dict[str, str] | None = ...
    ) -> None: ...
    def read_template_str(self, file: str) -> str: ...



"""Implementation of the base class of SIT. This class generates the
boilerplate code required to build the black box interface in SST
Interoperability Toolkit.

This class is purely virtual and therefore requires a child class to inherit
and implement its protected methods. The child class must implement the
following protected methods:
- _get_driver_inputs()
- _get_driver_outputs()
- _get_driver_defs()

The following public methods are inherited by the child classes and are not
to be overridden:
- set_ports(ports)
- generate_boilerplate()
- fixed_width_float_output(precision)
- disable_runtime_warnings(warnings)
"""

import math

from ..exceptions import ConfigException, PortException, SignalFormatException
from ..files import Paths
from ..render import TemplateRenderer


class HardwareDescriptionLanguage:
    def __init__(
        self,
        module_name,
        lib,
        ipc="sock",
        module_dir="",
        lib_dir="",
        desc="",
    ):
        """Constructor for the virtual base class SIT

        Initialize all member port variables and component variables. Only the
        following methods are public:
        `set_ports(ports)`
        `generate_boilerplate()`

        Parameters:
        -----------
        module_name : str
            SST element component and HDL module name
        lib : str
            SST element library name
        ipc : str (options: "sock", "zmq")
            method of IPC
        module_dir : str (default: "")
            directory of HDL module
        lib_dir : str (default: "")
            directory of SIT library
        desc : str (default: "")
            description of the SST model

        Raises:
        -------
        ConfigException
            unsupported IPC inputted
        """
        if ipc in ("sock", "zmq"):
            self.ipc = ipc
        else:
            raise ConfigException(f"{ipc} is not a supported IPC protocol")

        self.module_name = module_name
        self.lib = lib
        self.lib_dir = lib_dir
        self.desc = desc
        self.exec_cmd = ""

        self.precision = 0
        self.extra_libs = ""
        self.disable_warning = ""

        self.hdl_str = self.__class__.__name__.lower()

        self.width_macros = {}
        self.ports: dict[str, list[dict[str, int]]] = {
            "clock": [],
            "input": [],
            "output": [],
            "inout": [],
        }

        self.driver_buf_size = 0
        self.comp_buf_size = 0
        if self.ipc == "sock":

            # component attributes
            self.sig_type = "SITSocketBuffer"

        elif self.ipc == "zmq":

            # component attributes
            self.sig_type = "SITZMQBuffer"

        # shared attributes
        self.sender = self.receiver = "sit_buf"

        self.paths = Paths(self.hdl_str, module_dir)

        self.template = TemplateRenderer()

    def _get_driver_inputs(self):
        raise NotImplementedError()

    def _get_driver_outputs(self):
        raise NotImplementedError()

    def _get_driver_defs(self):
        raise NotImplementedError()

    def _compute_signal_buffer_len(self, signal_type, signal_len):
        raise NotImplementedError()

    def _generate_extra_files(self):
        raise NotImplementedError()

    @staticmethod
    def _sig_fmt(fmt, split_func, array, delim):
        """Format lists of signals based on fixed arguments

        Parameters:
        -----------
        fmt : str
            string format
        split_func : lambda or dict(str:str)
            map to split on the signals
        array : list(str)
            list of signals
        delim : str (default: ";n    ")
            delimiter

        Returns:
        --------
        str
            string formatted signals
        """
        return delim.join(fmt.format(**split_func(i)) for i in array)

    def _get_signal_width_from_macro(self, signal_type, signal_type_macro):
        """Get width of a signal type mapped in width_macros

        Parameters:
        -----------
        signal_type : str
            signal type of port
        signal_type_macro : str
            signal type macro of port

        Raises:
        -------
        SignalFormatException
            signal width macro not found

        Returns:
        --------
        value : str
            width of signal type
        """
        if signal_type_macro_value := self.width_macros.get(
            signal_type_macro, None
        ):
            return signal_type_macro_value

        raise SignalFormatException(
            f"Invalid macro in signal: {signal_type}"
        ) from None

    def _get_all_ports(self):
        """Flatten all types of ports into a single array

        Returns:
        --------
        generator
            all ports in a single array
        """
        return [i for sig in self.ports.values() for i in sig]

    def _get_input_ports(self):
        """Flatten all types of ports into a single array

        Returns:
        --------
        generator
            all ports in a single array
        """
        return self.ports["input"]

    def _get_output_ports(self):
        """Flatten all types of ports into a single array

        Returns:
        --------
        generator
            all ports in a single array
        """
        return self.ports["output"]

    @staticmethod
    def _get_num_digits(signal):
        """Compute the minimum number of digits required to hold signal data
        type width by calculating: `floor(log2(-1 + 2^x)/log2(10)) + 1`. This
        expression is equivalent to `ceil(x / log2(10))`

        Parameters:
        -----------
        signal : int
            signal data type width

        Returns:
        --------
        int
            number of digits for signal width
        """
        # log2(10) = 3.321928094887362
        return math.ceil(signal / 3.321928094887362)

    def __get_comp_defs(self):
        """Map definitions for the component format string

        Returns:
        --------
        dict(str:str)
            format mapping of template component string
        """
        self.comp_buf_size = sum(
            output_port["len"] for output_port in self.ports["output"]
        )
        if self.precision and self.precision > self.comp_buf_size - 2:
            print(
                f"Component buffer size increased from {self.comp_buf_size} to ",
                end="",
            )
            self.comp_buf_size += self.precision - 2
            print(f"{self.comp_buf_size} to include specified precision")

        return {
            "exec_cmd": self.exec_cmd,
            "lib_dir": self.lib_dir,
            "module_name": self.module_name,
            "lib": self.lib,
            "desc": self.desc,
            "ports": self._sig_fmt(
                fmt="""{{"{link}", "{desc}", {{"sst.Interfaces.StringEvent"}}}}""",
                split_func=lambda x: {
                    "link": self.module_name + x[0],
                    "desc": self.module_name + x[-1],
                },
                array=(("_din", " data in"), ("_dout", " data out")),
                delim=",\n" + " " * 8,
            ),
            "sig_type": self.sig_type,
            "buf_size": self.comp_buf_size,
            "sender": self.sender,
            "receiver": self.receiver,
        }

    def __generate_comp_str(self):
        """Generate the black box-model code based on methods used to format
        the template file

        Raises:
        -------
        TemplateFileNotFound
            boilerplate template file not found

        Returns:
        --------
        str
            boilerplate code representing the black box-model file
        """
        template = self.paths.read_template_str("comp")
        return self.template.render(template, self.__get_comp_defs())

    def __generate_driver_str(self):
        """Generate the black box-driver code based on methods used to format
        the template file

        Raises:
        -------
        TemplateFileNotFound
            boilerplate template file not found

        Returns:
        --------
        str
            boilerplate code representing the black box-driver file
        """
        template = self.paths.read_template_str("driver")
        return self.template.render(
            template,
            dict(
                inputs=self._get_driver_inputs(),
                outputs=self._get_driver_outputs(),
                **self._get_driver_defs(),
            ),
        )

    def set_template_paths(self, dir="", driver="", comp=""):
        self.paths.set_template_paths(dir, driver, comp)

    def set_gen_paths(self, dir="", driver="", comp=""):
        self.paths.set_gen_paths(dir, driver, comp)

    def set_width_macros(self, width_macros):
        """Set width macros provided by user

        width_macros : dict(str:[str|int]) (default: dict(None:None))
            mapping of signal width macros to their integer values. An HDL
            module may declare constants or user-inputted variables in their
            implementation to determine signal widths.
            The module can only know the values of those macros by utilizing
            this parameter.
            Example:
            `width_macros = {
                "ADDRESS_WIDTH": 16,
                "DATA_WIDTH": 16,
            }`"""
        self.width_macros = width_macros

    def fixed_width_float_output(self, precision):
        """Generate additional methods to handle ports with float signals

        Parameters:
        -----------
        precision : int
            level of precision for float signals

        Raises:
        -------
        AttributeError
            method not supported
        """
        self.precision = precision
        print("Adding fixed precision for float outputs")
        try:
            self._fixed_width_float_output()
        except AttributeError:
            raise AttributeError(
                f"fixed_width_float_output() not supported with {self.module_name}"
            )

    def disable_runtime_warnings(self, warnings):
        """Generate additional methods to disable driver runtime warnings

        Parameters:
        -----------
        warnings : str|list(str)
            runtime warning or list of runtime warnings to ignore

        Raises:
        -------
        AttributeError
            method not supported
        """
        if not isinstance(warnings, list):
            warnings = [warnings]
        for warning in warnings:
            try:
                self._disable_runtime_warnings(warning)
            except AttributeError:
                raise AttributeError(
                    f"disable_runtime_warnings() not supported with {self.module_name}"
                )

    def set_ports(self, ports):
        """Assign ports to their corresponding member lists

        Parameters:
        -----------
        ports : nested dict
            type-declared signals in the form:
                {"name": <SIGNAL NAME>, "type": <SIGNAL DATA TYPE>, "len": <SIGNAL LEN>}
            The current types of signals supported are:
                ("clock", "input", "output", "inout")

        Raises:
        -------
        SignalFormatException
            invalid signal format provided
        PortException
            invalid port type provided
        """
        for port_type, signals in ports.items():

            for signal in signals:

                if len(signal) != 3:
                    raise SignalFormatException(
                        "Invalid signal format"
                    ) from None

                signal["len"] = self._compute_signal_buffer_len(
                    signal_type=signal["type"], signal_len=signal["len"]
                )

                try:
                    self.ports[port_type].append(signal)

                except KeyError:
                    raise PortException(
                        f"{signal[port_type]} is an invalid port type"
                    ) from None

    def generate_boilerplate(self):
        """Provide a high-level interface to the user to generate both the
        components of the black box and dump them to their corresponding files

        Raises:
        -------
        PortException
            no ports were provided
        """
        if not len(self.ports):
            raise PortException(
                "No ports were set. Make sure to call set_ports() before generating files."
            )

        self.paths.get_gen("dir").mkdir(exist_ok=True)

        with open(self.paths.get_gen("driver"), "w") as driver_file:
            driver_file.write(self.__generate_driver_str())

        with open(self.paths.get_gen("comp"), "w") as comp_file:
            comp_file.write(self.__generate_comp_str())

        try:
            self._generate_extra_files()
        except NotImplementedError:
            pass



"""Implementation of the PyRTL class

This class inherits from the HardwareDescriptionLanguage base class and implements its own methods of
parsing, modifying and generating boilerplate code for its specific paradigms.
"""
from typing import Literal

from ..it import HardwareDescriptionLanguage

class PyRTL(HardwareDescriptionLanguage):
    def __init__(
        self,
        module_name: str,
        lib: str,
        ipc: Literal["sock", "zmq"] = ...,
        module_dir: str = ...,
        lib_dir: str = ...,
        desc: str = ...,
    ) -> None: ...
    def _compute_signal_buffer_len(
        self, signal_type: str, signal_len: int
    ) -> int: ...
    def _get_driver_inputs(self) -> str: ...
    def _get_driver_outputs(self) -> str: ...
    def _get_driver_defs(self) -> dict[str, str]: ...



class ConfigFile:
    def generate_linkwrapper_binding(
        self,
        module_name: str,
        lib: str,
        input_ports: list[dict[str, str | int]],
    ) -> str: ...
    def generate_comps_func(self, module_name: str) -> str: ...



import sys

if __name__ == "__main__":

    with open(sys.argv[-1]) as mem_dump_file:

        data = mem_dump_file.read().split("\n")[:-1]

        try:
            data_str = "".join([chr(int(x, 2)) for x in data])
        except ValueError:
            data_str = "".join([chr(int(x)) for x in data])

        raise SystemExit(0) if data_str == "HELLO WORLD" else SystemExit(1)



import pathlib

from .exceptions import TemplateFileNotFound


class Paths:
    def __init__(self, hdl_str, module_dir_path):

        self.hdl_str = hdl_str

        self.__module_dir = pathlib.Path(module_dir_path)

        self.__template_paths = {}
        self.__template_paths["dir"] = (
            pathlib.Path(__file__).parent / "hdl" / hdl_str / "templates"
        ).resolve()
        self.__template_paths["driver"] = (
            self.__template_paths["dir"] / "driver"
        )
        self.__template_paths["comp"] = (
            pathlib.Path(__file__).parent / "hdl" / "templates" / "comp"
        )

        self.__gen_paths = {}
        self.__gen_paths["dir"] = pathlib.Path("gen").resolve()

    def get_module_dir(self):

        return self.__module_dir

    def get_template(self, key):

        return self.__template_paths[key]

    def get_gen(self, key):

        return self.__gen_paths[key]

    def get_paths(self):

        return self.__module_dir, self.__template_paths, self.__gen_paths

    def set_template_paths(self, dir="", driver="", comp=""):

        if dir:
            self.__template_paths["dir"] = pathlib.Path(dir)

        if driver:
            self.__template_paths["driver"] = pathlib.Path(driver)

        if comp:
            self.__template_paths["comp"] = pathlib.Path(comp)

    def set_gen_paths(self, dir="", driver="", comp=""):

        if dir:
            self.__gen_paths["dir"] = pathlib.Path(dir)

        if driver:
            self.__gen_paths["driver"] = pathlib.Path(driver)

        if comp:
            self.__gen_paths["comp"] = pathlib.Path(comp)

    def set_driver_path(self, driver_file):

        self.__gen_paths["driver"] = self.__gen_paths["dir"] / driver_file

    def set_comp_path(self, comp_file):

        self.__gen_paths["comp"] = self.__gen_paths["dir"] / comp_file

    def set_extra_file_paths(self, templates, gens=None):

        if gens and templates.keys() == gens.keys():

            for file_key in templates:
                self.__gen_paths[file_key] = (
                    self.__gen_paths["dir"] / gens[file_key]
                )
                self.__template_paths[file_key] = (
                    self.__template_paths["dir"] / templates[file_key]
                )

        else:
            for file_key in templates:
                self.__gen_paths[file_key] = (
                    self.__gen_paths["dir"] / templates[file_key]
                )
                self.__template_paths[file_key] = (
                    self.__template_paths["dir"] / templates[file_key]
                )

    def read_template_str(self, file):

        template_file = self.__template_paths[file]
        if template_file.exists():
            with open(template_file) as fp:
                return fp.read()

        raise TemplateFileNotFound(
            f"Component boilerplate template file: '{template_file}' not found"
        )



"""Implementation of the SystemC class

This class inherits from the HardwareDescriptionLanguage base class and implements its own methods of
parsing, modifying and generating boilerplate code for its specific paradigms.
"""
import warnings

from .. import HardwareDescriptionLanguage


class SystemC(HardwareDescriptionLanguage):
    def __init__(
        self,
        module_name,
        lib,
        ipc="sock",
        module_dir="",
        lib_dir="",
        desc="",
    ):
        """Constructor for SystemC

        Parameters:
        -----------
        ipc : str (options: "sock", "zmq")
            method of IPC
        module : str
            SST element component and HDL module name
        lib : str
            SST element library name
        module_dir : str (default: "")
            directory of HDL module
        lib_dir : str (default: "")
            directory of library
        desc : str (default: "")
            description of the SST model
        driver_template_path : str (default: "")
            path to the black box-driver boilerplate
        component_template_path : str (default: "")
            path to the black box-model boilerplate
        """
        super().__init__(
            module_name=module_name,
            lib=lib,
            ipc=ipc,
            module_dir=module_dir,
            lib_dir=lib_dir,
            desc=desc,
        )

        if self.ipc == "sock":

            # driver attributes
            self.sig_type = "SITSocketBuffer"

        elif self.ipc == "zmq":

            # driver attributes
            self.sig_type = "SITZMQBuffer"

        self.exec_cmd = """char* args[] = {&m_proc[0u], &m_ipc_port[0u], nullptr};

            m_output.verbose(
                CALL_INFO, 1, 0, "\\033[35m[FORK] %s\\033[0m\\n\", cmd.c_str()
            );"""

        self.paths.set_driver_path(f"{self.module_name}_driver.cpp")
        self.paths.set_comp_path(f"{self.module_name}_comp.cpp")
        self.__data_out_str = ""

    def __extract_int(self, signal_type):
        """Extract integers from signal string
        This is a closure method to keep scope local

        Returns:
        --------
        int
            integers found in signal string
        """
        signal_type_macro = ""
        parse_flag = False
        for c in signal_type:

            if c == "<":
                parse_flag = True
                continue
            elif c == ">":
                break

            if parse_flag:
                signal_type_macro += c

        if signal_type_macro.isdigit():
            return int(signal_type_macro)

        return int(
            self._get_signal_width_from_macro(signal_type, signal_type_macro)
        )

    def _compute_signal_buffer_len(self, signal_type, signal_len):
        """Parse the type and computes its size from the signal

        Parameters:
        -----------
        signal : str
            signal definition

        Returns:
        --------
        int
            signal width
        """
        # SystemC member data types
        if "sc" in signal_type:

            if any(x in signal_type for x in ("bit", "logic")):

                if not (signal_len == 1):
                    warnings.warn(
                        f'Length of "{signal_type}" types will always be 1'
                    )
                return 1

            if not (signal_len == -1):
                warnings.warn(
                    f'Length of "{signal_type}" type lengths will be automatically computed. Explicit values will be ignored.'
                )

            signal_type_int = self.__extract_int(signal_type)
            if any(x in signal_type for x in ("bv", "lv")):
                return signal_type_int

            elif any(x in signal_type for x in ("int", "uint")):
                return self._get_num_digits(signal_type_int)

        # boolean signals
        elif signal_type == "bool":

            if not (signal_len == 1):
                warnings.warn('Length of "bool" types will always be 1')

            return 1

        # default C++ data types
        else:
            return signal_len

    def _get_driver_outputs(self):
        """Generate output bindings for both the components in the black box

        Returns:
        --------
        str
            snippet of code representing output bindings
        """
        if len(self._get_output_ports()) > 1:
            self.extra_libs += "#include <sst/sit/bufwidth.hpp>\n"
            self.__data_out_str = "std::string _data_out_str;"
            return self._sig_fmt(
                fmt="""_data_out_str = align_buffer_width(std::to_string({sig}.read()), {len});
        _data_out << _data_out_str""",
                split_func=lambda x: {"sig": x["name"], "len": x["len"]},
                array=self._get_output_ports(),
                delim=";\n" + " " * 8,
            )

        else:
            return self._sig_fmt(
                fmt="_data_out << {precision}{sig}",
                split_func=lambda x: {
                    "sig": x["name"],
                    "precision": f"std::fixed << std::setprecision({self.precision}) << "
                    if self.precision and x["type"] == "float"
                    else "",
                },
                array=self._get_output_ports(),
                delim=";\n" + " " * 8,
            )

    def _get_driver_inputs(self):
        """Generate input bindings for the driver.

        Returns:
        --------
        str
            snippet of code representing input bindings
        """
        fmt = "{sig} = std::sto{type}(_data_in.substr({sp}, {sl}));"
        start_pos = 1
        clock_fmt = "{sig} = std::stol(_data_in.substr({sp}, {sl})) % 2;"
        driver_inputs = []

        # input_port = (INPUT NAME, INPUT TYPE)
        for input_port in self._get_input_ports():
            driver_inputs.append(
                fmt.format(
                    type="f" if input_port["type"] == "float" else "l",
                    sp=start_pos,
                    sl=str(input_port["len"]),
                    sig=input_port["name"],
                )
            )
            start_pos += input_port["len"]

        if self.ports["clock"]:
            for clock_port in self.ports["clock"]:
                driver_inputs.append(
                    clock_fmt.format(
                        sp=start_pos,
                        sl=clock_port["len"],
                        sig=clock_port["name"],
                    )
                )
                start_pos += int(clock_port["len"])

        self.driver_buf_size = start_pos
        return ("\n" + " " * 8).join(driver_inputs)

    def __get_driver_port_defs(self):
        """Generate port definitions for the black box-driver

        Returns:
        --------
        str
           string format of driver port definitions
        """
        return "sc_signal" + ";\n    sc_signal".join(
            f"""<{i["type"]}> {i["name"]}""" for i in self._get_all_ports()
        )

    def __get_driver_bindings(self):
        """Generate port bindings for the black box-driver

        Returns:
        --------
        str
            snippet of code representing port bindings
        """
        return self._sig_fmt(
            fmt="DUT.{sig}({sig})",
            split_func=lambda x: {"sig": x["name"]},
            array=self._get_all_ports(),
            delim=";\n    ",
        )

    def _get_driver_defs(self):
        """Map definitions for the SystemC driver format string

        Returns:
        --------
        dict(str:str)
            format mapping of template SystemC driver string
        """
        return {
            "extra_libs": self.extra_libs,
            "module_dir": self.paths.get_module_dir().resolve(),
            "lib_dir": self.lib_dir,
            "module_name": self.module_name,
            "disable_warning": self.disable_warning,
            "port_defs": self.__get_driver_port_defs(),
            "bindings": self.__get_driver_bindings(),
            "multi_outputs": self.__data_out_str,
            "sender": self.sender,
            "receiver": self.receiver,
            "buf_size": self.driver_buf_size,
            "sig_type": self.sig_type,
        }

    def _fixed_width_float_output(self):
        """Add <iomanip> library to handle ports with float signals"""
        self.extra_libs += "#include <iomanip>\n"

    def _disable_runtime_warnings(self, warning):
        """Add sc_report_handler to disable driver runtime warnings

        Parameters:
        -----------
        warning : str
            runtime warning to ignore
        """
        self.disable_warning += (
            f"sc_report_handler::set_actions({warning}, SC_DO_NOTHING);"
        )



import os
import subprocess

class Commands:
    @staticmethod
    def sanitize(cmd: subprocess.CompletedProcess[bytes]) -> str: ...
    def run_cmd(
        self, cmd: list[str], stdout: int | None = ...
    ) -> subprocess.CompletedProcess[bytes]: ...
    def __get_sst_config(self) -> str: ...
    def __get_prefix(self, sst_config: str) -> os.PathLike: ...
    def __get_include_dir(self, sst_config: str) -> os.PathLike: ...
    def get_dest_dir(self) -> os.PathLike: ...

class LibraryManager:
    def __init__(self) -> None: ...
    def is_installed(self) -> bool: ...
    def get_dest_dir(self) -> os.PathLike: ...
    def install(self) -> None: ...



"""Implementation of the Verilog class

This class inherits from the HardwareDescriptionLanguage base class and implements its
own methods of parsing, modifying and generating boilerplate code for its
specific paradigms.
"""
from .. import HardwareDescriptionLanguage
from ...exceptions import SignalFormatException


class Verilog(HardwareDescriptionLanguage):
    def __init__(
        self,
        module_name,
        lib,
        ipc="sock",
        module_dir="",
        lib_dir="",
        desc="",
    ):
        """Constructor for Verilog HardwareDescriptionLanguage

        Parameters:
        -----------
        module : str
            SST element component and HDL module name
        lib : str
            SST element library name
        ipc : str (options: "sock", "zmq")
            method of IPC
        module_dir : str (default: "")
            directory of HDL module
        lib_dir : str (default: "")
            directory of HardwareDescriptionLanguage library
        desc : str (default: "")
            description of the SST model
        """
        super().__init__(
            module_name=module_name,
            lib=lib,
            ipc=ipc,
            module_dir=module_dir,
            lib_dir=lib_dir,
            desc=desc,
        )

        if self.ipc == "sock":

            # driver attributes
            self.driver_ipc = "socket"
            self.driver_bind = (
                "_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)"
            )
            self.send = "sendall"
            self.connect = "connect"

        elif self.ipc == "zmq":

            # driver attributes
            self.driver_ipc = "zmq"
            self.driver_bind = """context = zmq.Context()
    _sock = context.socket(zmq.REQ)"""
            self.send = "send"
            self.connect = "bind"

        self.exec_cmd = """m_output.verbose(
                CALL_INFO, 1, 0, "\\033[35m[FORK] %s\\033[0m\\n\", cmd.c_str()
            );
            char* args[8];
            int i = 0;
            args[i] = std::strtok(&cmd[0u], " ");

            while (args[i] != nullptr) {
                args[++i] = strtok(nullptr, " ");
            }
            args[i] = nullptr;"""

        self.paths.set_driver_path(f"{self.module_name}_driver.py")
        self.paths.set_comp_path(f"{self.module_name}_comp.cpp")
        self.paths.set_extra_file_paths({"makefile": "Makefile.config"})

    def _compute_signal_buffer_len(self, signal_type, signal_len):
        """Parse the type and computes its size from the signal

        Parameters:
        -----------
        signal : str
            signal definition

        Returns:
        --------
        int
            signal width
        """

        match signal_type:

            case "int":
                return self._get_num_digits(signal_len)

            case "bit":
                return signal_len

            case _:
                raise SignalFormatException(
                    f'Invalid signal type "{signal_type}"'
                ) from None

    def _get_driver_outputs(self):
        """Generate output bindings for both the components in the black box

        Returns:
        --------
        str
            snippet of code representing output bindings
        """
        return self._sig_fmt(
            fmt="str(dut.{sig}.value{type})",
            split_func=lambda x: {
                "sig": x["name"],
                "type": (".integer" if x["type"] == "int" else ""),
            },
            array=self._get_output_ports(),
            delim="\n" + (" " * 12) + "+ ",
        )

    def _get_driver_inputs(self):
        """Generate input bindings for the driver.

        Returns:
        --------
        str
            snippet of code representing input bindings
        """
        fmt = "dut.{sig}.value = int(signal[{sp}:{sl}])"
        start_pos = 0
        driver_inputs = []

        for input_port in self._get_input_ports():
            driver_inputs.append(
                fmt.format(
                    sp=start_pos,
                    sl=str(input_port["len"] + start_pos),
                    sig=input_port["name"],
                )
            )
            start_pos += input_port["len"]

        if self.ports["clock"]:
            clock_fmt = "dut.{sig}.value = int(signal[{sp}:{sl}]) % 2"

            for clock_port in self.ports["clock"]:
                driver_inputs.append(
                    clock_fmt.format(
                        sp=start_pos,
                        sl=str(input_port["len"] + start_pos),
                        sig=clock_port["name"],
                    )
                )
                start_pos += int(clock_port["len"])

        self.driver_buf_size = start_pos + 1
        return ("\n" + " " * 8).join(driver_inputs)

    def _get_driver_defs(self):
        """Map definitions for the Verilog driver format string

        Returns:
        --------
        dict(str:str)
            format mapping of template Verilog driver string
        """
        return {
            "ipc": self.driver_ipc,
            "driver_bind": self.driver_bind,
            "connect": self.connect,
            "send": self.send,
            "module_name": self.module_name,
            "buf_size": self.driver_buf_size,
        }

    def _generate_extra_files(self):

        template = self.paths.read_template_str("makefile")
        template_str = self.template.render(
            template,
            dict(
                module_name=self.module_name,
                module_dir=self.paths.get_module_dir().resolve(),
            ),
        )

        with open(self.paths.get_gen("makefile"), "w") as makefile:
            makefile.write(template_str)
        print(f"Dumped Makefile to '{self.paths.get_gen('makefile')}'")



import os
from typing import Callable, Literal, final

class HardwareDescriptionLanguage:
    def __init__(
        self,
        module_name: str,
        lib: str,
        ipc: Literal["sock", "zmq"],
        module_dir: str = ...,
        lib_dir: str = ...,
        desc: str = ...,
    ) -> None: ...
    def __get_comp_defs(self) -> dict[str, str]: ...
    def __generate_comp_str(self) -> str: ...
    def __generate_driver_str(self) -> str: ...
    def _get_driver_inputs(self) -> str: ...
    def _get_driver_outputs(self) -> str: ...
    def _get_driver_defs(self) -> str: ...
    def _compute_signal_buffer_len(
        self, signal_type: str, signal_len: int
    ) -> int: ...
    @staticmethod
    def _sig_fmt(
        fmt: str,
        split_func: Callable[[dict[str, str]], dict[str, str]],
        array: list[dict[str, str | int]],
        delim: str,
    ) -> str: ...
    def _get_signal_width_from_macro(
        self, signal_type: str, signal_type_macro: str
    ) -> str: ...
    def _get_all_ports(self) -> list[dict[str, str | int]]: ...
    def _get_input_ports(self) -> list[dict[str, str | int]]: ...
    def _get_output_ports(self) -> list[dict[str, str | int]]: ...
    @staticmethod
    def _get_num_digits(signal: int) -> int: ...
    def set_template_paths(
        self,
        template_dir_path: os.PathLike = ...,
        driver_template_path: os.PathLike = ...,
        component_template_path: os.PathLike = ...,
    ) -> None: ...
    def set_gen_paths(
        self,
        gen_dir_path: os.PathLike = ...,
        driver_path: os.PathLike = ...,
        comp_path: os.PathLike = ...,
    ) -> None: ...
    @final
    def fixed_width_float_output(self, precision: int) -> None: ...
    @final
    def disable_runtime_warnings(self, warnings: str | list[str]) -> None: ...
    @final
    def set_width_macros(self, width_macros: dict[str, int]) -> None: ...
    @final
    def set_ports(
        self,
        ports: dict[str, str | dict[str, str | list[dict[str, str | int]]]],
    ) -> None: ...
    @final
    def generate_boilerplate(self) -> None: ...



class HardwareDescriptionLanguage:
    def __init__(
        self,
        config_data: dict[
            str, str | dict[str, str | list[dict[str, str | int]]]
        ],
    ) -> None: ...
    def generate_boilerplate(self): ...
    def set_config_data(
        self,
        config_data: dict[
            str, str | dict[str, str | list[dict[str, str | int]]]
        ],
    ) -> None: ...
    def dump_summary(self) -> None: ...
    def generate_config_file(self) -> None: ...
    def __set_hdl_str(self, hdl_str: str) -> None: ...

def run() -> None: ...



"""Custom exceptions for the SIT base class"""


class SITError(Exception):
    """Base exception class"""

    pass


class ConfigException(SITError):
    """Exceptions for invalid or unsupported configuration, including IPC method and HDL"""

    pass


class PortException(SITError):
    """Exceptions for invalid port types and empty ports"""

    pass


class SignalFormatException(SITError):
    """Exceptions for invalid signal format"""

    pass


class TemplateFileNotFound(SITError):
    """Exceptions for missing or invalid path to template file"""

    pass
